"""
This module is a collection of tools that are useful
to manage modules and odoo environment.

The submodule modules can be used to manipulate and
search existing module in a given environment.

The module ``odoo_tools.odoo`` gives a simple api
to work with an odoo environment. It's an api
that consume the collection of function in other
modules of this package.
"""



"""
Environment Variables
=====================
"""
from os import environ
from .utils import from_csv


class EnvironmentVariables(object):
    """
    EnvironmentVariables parser

    Attributes:

    ODOO_BASE_PATH: The base path where odoo is installed


    ODOO_RC: Location of the odoo.cfg file

    ODOO_STRICT_MODE: Run odoo in Strict Mode

    ODOO_EXCLUDED_PATHS (list): Excluded paths will not be looked into when
        searching modules

    ODOO_EXTRA_PATHS (list): Extra paths are paths to look for modules other
        than the default ones.

    ODOO_EXTRA_APT_PACKAGES: Extra apt packages to install other than the one
        introspected in modules.

    ODOO_DISABLED_MODULES: Odoo modules that shouldn't exist. Those would get
        removed from the addons paths.

    MASTER_PASSWORD: Master password that should be set instead of the default
        one being automatically generated randomly.

    SHOW_MASTER_PASSWORD: Log the master password in the logs.

    SKIP_PIP: Skip the installation of pip modules.

    SKIP_SUDO_ENTRYPOINT: Skip the sudo entrypoint. The entry point is used
        mainly to setup things that require higher access like installing apt
        packages.

    SKIP_POSTGRES_WAIT: Tells to skip waiting for postgress to be up and
        running

    I_KNOW_WHAT_IM_DOING: This environment variables allows it to set the
        PGPASSWORD in the environment variables. Otherwise, the entrypoint
        will fail to let odoo starts unless you specify this environment
        variable as TRUE.  The reason behind that is that it's generally not
        a good idea to set credentials in environment variables. Those values
        can be leaked easily in logs from the environment running the container
        or even in logs displaying environment variables anywhere. This
        should only be used in dev environment or test environment in which
        credentials are not that important.

    DEPLOYMENT_AREA: Deployment Area is just an environment variable that can
        be set to determine if it's a production environment or test
        environment.

    ODOO_VERSION: The odoo version being currently used.

    RESET_ACCESS_RIGHTS: Reset access rights ensure that files in the home
        directory of the odoo user is set to the right user. This shouldn't
        be necessary in most cases.

    DEPLOYMENT_AREA: Deployment Area is just an environment variable that can
        be set to determine if it's a production environment or test
        environment.

    APT_INSTALL_RECOMMENDS: Define if you want the sudo_entrypoint to call
        apt-get with --no-install-recommends. By default, the entrypoint will
        use --no-install-recommends, but if this environment variable is set
        to TRUE. It will not add this parameter to apt-get install.

    ODOO_REQUIREMENTS_FILE: The path in which requirements get stored during
        installation of pip packages required by odoo addons.

    USE_ODOO_LOGGER: Tell the app to initialize the odoo logger instead of
        using the default one.

    PACKAGE_MAP_FILE: File storing a map of {module_name: module_renamed} to
        rename modules that shouldn't be used as is.
    """
    def __init__(self):
        self.ODOO_BASE_PATH = environ.get('ODOO_BASE_PATH')
        self.ODOO_RC = environ.get('ODOO_RC')
        self.ODOO_STRICT_MODE = environ.get('ODOO_STRICT_MODE')
        self.ODOO_EXCLUDED_PATHS = from_csv(
            environ.get('ODOO_EXCLUDED_PATHS', ''),
            set
        )
        self.ODOO_EXTRA_PATHS = from_csv(
            environ.get('ODOO_EXTRA_PATHS'),
            set
        )
        self.ODOO_EXTRA_APT_PACKAGES = from_csv(
            environ.get('ODOO_EXTRA_APT_PACKAGES') or
            environ.get('EXTRA_APT_PACKAGES'),
            set
        )
        self.ODOO_DISABLED_MODULES = from_csv(
            environ.get('ODOO_DISABLED_MODULES'),
            set
        )

        self.MASTER_PASSWORD = environ.get('MASTER_PASSWORD')
        self.SHOW_MASTER_PASSWORD = environ.get('SHOW_MASTER_PASSWORD')

        self.SKIP_PIP = environ.get('SKIP_PIP')
        self.SKIP_SUDO_ENTRYPOINT = environ.get('SKIP_SUDO_ENTRYPOINT')
        self.SKIP_POSTGRES_WAIT = environ.get('SKIP_POSTGRES_WAIT')
        self.I_KNOW_WHAT_IM_DOING = environ.get('I_KNOW_WHAT_IM_DOING')

        self.ALLOW_DANGEROUS_SETTINGS = environ.get(
            'I_KNOW_WHAT_IM_DOING',
            environ.get(
                'ALLOW_DANGEROUS_SETTINGS',
                ''
            )
        ).lower() == 'true'

        #: Deployment Area is just an environment variable that can be set
        #: to determine if it's a production environment or test environment.
        self.DEPLOYMENT_AREA = environ.get('DEPLOYMENT_AREA')

        #: The odoo version being currently used.
        self.ODOO_VERSION = environ.get('ODOO_VERSION')

        self.RESET_ACCESS_RIGHTS = environ.get('RESET_ACCESS_RIGHTS', '')

        self.APT_INSTALL_RECOMMENDS = environ.get('APT_INSTALL_RECOMMENDS')

        self.ODOO_REQUIREMENTS_FILE = environ.get('ODOO_REQUIREMENTS_FILE')

        self.USE_ODOO_LOGGER = environ.get('USE_ODOO_LOGGER')

        self.PACKAGE_MAP_FILE = environ.get('PACKAGE_MAP_FILE')

        self.REQUIREMENTS_FILE_PATH = environ.get('REQUIREMENTS_FILE_PATH')



import six
from six import ensure_binary, ensure_text
import shutil
import logging
from ast import literal_eval
import tempfile
from zipfile import ZipFile

from ..compat import Path
from ..modules.render import render_description_str
from ..modules.translate import PoFileWriter, PoFileReader
from ..exceptions import ArgumentError

_logger = logging.getLogger(__name__)


class CompanySpec(object):
    def __init__(self, country_code):
        self.country_code = country_code


def try_parse_manifest(data):
    return literal_eval(data)


def try_compile_manifest(data):
    code = compile(
        data,
        '__manifest__.py',
        'eval'
    )
    return eval(code, {}, {})


def get_translation_filename(language, module):
    """
    Get the filename for translation file.

    When no language is provided to translate a module,
    the name of the file will the name of the module as
    a pot file.

    For example, exporting the pot file for module bus would be
    exported by calling:

    ..code:: python

    get_translation_filename(None, 'bus')
    >> 'bus.pot'

    When exporting the translation file for fr_FR language for
    the module bus. It will return the family of the language
    as the langage and dialect are the same.

    ..code:: python

    get_translation_filename('fr_FR', 'bus')
    >> 'fr.po'

    If the language and dialect are different it returns the language
    locale as is.

    ..code:: python
    get_translation_filename('fr_CA', 'bus')
    >> 'fr_CA.po'

    Args:

        language (str|None): locale to export

        module (str): Name of the module to export

    Returns:

        str: The filename of the translation file based
            on the arguments.
    """
    if language and '_' in language:
        family, dialect = language.lower().split('_')
        if family == dialect:
            language = family

    filename = (
        '{}.po'.format(language)
        if language
        else '{}.pot'.format(module)
    )
    return filename


class Manifest(object):
    """
    Manifest Object

    The manifest object contains data stored in odoo/openerp manifests.
    """

    properties = {
        "_attrs",
        "path",
        "_manifest_file"
    }

    defaults = {
        "installable": True,
        "application": False,
        "depends": [],
        "demo": [],
        "data": [],
        "version": "0.0.0",
        "external_dependencies": dict
    }

    def __init__(self, path, attrs=None, manifest_file=None):
        """
        Initialize manifest objects. All attributes are stored into
        `_attrs`.
        """
        if attrs is None:
            attrs = {}

        self._attrs = attrs
        self.path = Path(path)

        if 'technical_name' not in attrs:
            attrs['technical_name'] = self.path.name

        self._manifest_file = manifest_file or (self.path / '__manifest__.py')

        self.set_defaults()

    def set_defaults(self):
        """
        Set some defaults when initializing an object
        """
        for key, value in Manifest.defaults.items():
            if key not in self._attrs:
                if callable(value):
                    self._attrs[key] = value()
                else:
                    self._attrs[key] = value

    def __getattr__(self, name, default=None):
        return self._attrs.get(name, default)

    def __setattr__(self, name, value):
        if (
            name in Manifest.properties
            # name in Manifest.defaults.keys()
        ):
            super(Manifest, self).__setattr__(name, value)
        else:
            self._attrs[name] = value

    def __lt__(self, other):
        return self.path.name < other.path.name

    def __eq__(self, other):
        if isinstance(other, Manifest):
            return self.path == other.path
        elif isinstance(other, str):
            return self.path.name == Path(other).name
        else:
            return False

    def __contains__(self, value):
        return value in self._attrs

    def __hash__(self):
        return hash(self.path.name)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return "Manifest({path}, {attrs})".format(
            path=repr(self.path),
            attrs=repr(self._attrs)
        )

    def values(self):
        """
        Return manifest values
        """
        return self._attrs

    def set_attribute(self, attribute, value):
        if len(attribute) == 0:
            raise ArgumentError(
                'The attribute must have at least one attribute name.'
            )
        cur_vals = self._attrs

        last_attribute = attribute[-1]

        for attr in attribute[:-1]:
            if attr not in cur_vals:
                cur_vals[attr] = {}
            cur_vals = cur_vals[attr]
        else:
            cur_vals[last_attribute] = value

    def static_assets(self):
        addons_path = self.path.parent

        def recurse_search(cur_dir):
            files = []
            for file in cur_dir.iterdir():
                if file.is_dir():
                    files += recurse_search(file)
                else:
                    static_path = file.relative_to(addons_path)
                    files.append((static_path, file))
            return files

        if (self.path / "static").exists():
            return recurse_search(self.path / "static")
        else:
            return []

    @classmethod
    def from_path(klass, manifest, render_description=False):
        """
        Loads the manifest from a given path.

        It automatically define the module name to the path name of the
        folder in which the manifest is located to those attributes:

        - name
        - technical_name

        A custom attribute `description_html` can be generated to store
        the rendered html description.

        Args:
            manifest (Path): The path of the manifest or module.
            render_description (bool): Set to True if you need to render
              the description.

        Returns:
            Manifest: The manifest that was loaded.
        """
        if not manifest.name.endswith('py'):
            module_path = manifest
            odoo_manifest = module_path / '__manifest__.py'
            erp_manifest = module_path / '__openerp__.py'

            manifest = None
            if odoo_manifest.exists():
                manifest = odoo_manifest
            if not manifest and erp_manifest.exists():
                manifest = erp_manifest
        else:
            module_path = manifest.parent

        with manifest.open('rb') as fin:
            manifest_data = ensure_text(fin.read() or "")
            manifest_data = manifest_data.replace('\ufeff', '')

        parsers = [
            try_parse_manifest,
            try_compile_manifest,
            # try_parse_manifest_encoded,
            # try_compile_manifest_encoded
        ]

        last_error = None
        for parser in parsers:
            try:
                data = parser(manifest_data)
                break
            except Exception as exc:
                last_error = exc
        else:
            data = {}
            if last_error is not None:
                _logger.error(
                    "Cannot parse manifest: {}".format(
                        manifest,
                    ),
                )
                raise last_error

        if 'name' not in data:
            data['name'] = module_path.name

        data['technical_name'] = module_path.name

        if render_description:
            data['description_html'] = render_description_str(
                module_path,
                data.get('description', '')
            )

        man = Manifest(module_path, attrs=data, manifest_file=manifest)

        return man

    def save(self):
        obj = {}
        for key, value in self._attrs.items():
            if key in ['description_html', 'technical_name']:
                continue

            if isinstance(value, Path):
                obj[key] = str(value)
            else:
                obj[key] = value

        if not self._manifest_file.parent.exists():
            self._manifest_file.parent.mkdir(parents=True)

        with self._manifest_file.open('w') as fout:
            fout.write(repr(obj))

    def requirements(self, package_map=None):
        """
        Return a list of packages required by module.
        """
        packages = set()

        for package in self.external_dependencies.get('python', []):
            package_lower = package.lower()

            if package_map and package_lower in package_map:
                package_name = package_map[package.lower()]
                if package_name:
                    packages.add(package_name)
            else:
                packages.add(package)

        return packages

    def disable(self):
        self.installable = False

    def remove(self):
        """
        Remove the module from the file system
        """
        if self.path.exists():
            shutil.rmtree(self.path, ignore_errors=True)

    def files(self):
        for file in self.path.glob('**/*'):
            if file.is_dir():
                continue
            if file.name.endswith('.pyc'):
                continue

            yield file

    def package(self):
        outfile = tempfile.NamedTemporaryFile()

        root_folder = self.path.parent
        zipfile = ZipFile(outfile, 'w')

        for file in self.files():
            zip_filename = file.relative_to(root_folder)

            with file.open('rb') as fin:
                with zipfile.open(str(zip_filename), mode='w') as fout:
                    fout.write(fin.read())

        return outfile

    def export_translations(self, db, languages):

        translations = db.export_translation_terms(
            languages,
            [self.technical_name]
        )

        for language, module, rows in translations:
            filename = get_translation_filename(language, module)
            trans_path = self.path / 'i18n' / filename
            trans_path.parent.mkdir(parents=True, exist_ok=True)

            _logger.info(
                "Exporting translation %s of module %s to %s",
                language,
                module,
                trans_path
            )

            if trans_path.exists():
                origin_po_file = PoFileReader(trans_path.open('rb'))
            else:
                with trans_path.open("w+") as buffer:
                    buffer.write("")

                with trans_path.open('rb') as buffer:
                    origin_po_file = PoFileReader(buffer)

            with trans_path.open('wb+') as buffer:
                po_writer = PoFileWriter(
                    buffer,
                    language,
                    pofile=origin_po_file.pofile
                )

                po_writer.add_entries(rows)

                po_writer.write()



import six
import logging

from .db import DbApi
from ..entrypoints import execute_entrypoint
from ..configuration.odoo import (
    OfficialRelease,
    GitRelease
)


_logger = logging.getLogger(__name__)


class MangementApi(object):
    def __init__(self, environment):
        self.environment = environment
        self._initialized = False

    def db(self, database):
        return DbApi(
            self,
            database
        )

    @property
    def config(self):
        try:
            from odoo.tools import config
            return config
        except Exception:
            pass

        raise SystemError("Odoo doesn't seems to be installed")

    def initialize_odoo(self):
        self._initialized = True
        execute_entrypoint("odoo_tools.manage.before_config", self)
        execute_entrypoint("odoo_tools.manage.initialize_odoo", self)
        execute_entrypoint("odoo_tools.manage.after_config", self)

    def install_odoo(
        self,
        version,
        release=None,
        ref=None,
        repo='https://github.com/odoo/odoo.git',
        opts=None
    ):
        if hasattr(opts, 'cache'):
            cache = opts.cache
        else:
            cache = None

        if release:
            installer = OfficialRelease(
                version,
                release,
                options=opts,
                cache=cache
            )
        else:
            installer = GitRelease(
                version,
                repo,
                ref or version,
                options=opts,
                cache=cache
            )

        with installer:
            if installer.need_update():
                installer.fetch()
                installer.checkout()
                installer.install()

    def packages(self):
        package_list = set()
        paths = self.environment.addons_paths()

        _logger.info(
            "Looking up for packages in %s",
            paths
        )

        for addons_path in paths:
            for packages in addons_path.glob('**/apt-packages.txt'):
                _logger.info("Installing packages from %s", packages)

                with packages.open('r') as pack_file:
                    lines = [
                        six.ensure_text(line).strip()
                        for line in pack_file
                        if six.ensure_text(line).strip()
                    ]

                    package_list.update(set(lines))

        context = self.environment.context

        _logger.info(
            "Adding extra packages %s",
            context.extra_apt_packages
        )

        for package in context.extra_apt_packages:
            package_list.add(
                six.ensure_text(package).strip()
            )

        return package_list



"""
Search
======

The module search regroup a series of functions used to search
modules and handle their dependencies. This collection of functions,
can be used to determine which folder contains installable modules
and it can also be used to predetermine which module would require
to be installed if a given module was installed.
"""
import logging

from odoo_tools.compat import Path, module_path
from ..api.objects import Manifest


import pkg_resources

_logger = logging.getLogger(__name__)


def filter_installable(manifest):
    return manifest.installable


def filter_noninstallable(manifest):
    return not filter_installable(manifest)


def filter_python_dependencies(manifest):
    if (
        manifest.external_dependencies and
        manifest.external_dependencies.get('python')
    ):
        return True
    else:
        return False


def get_filter(filter_names):
    if not filter_names:
        return lambda module: True

    filters = []

    if 'installable' in filter_names:
        filters.append(filter_installable)

    if 'non_installable' in filter_names:
        filters.append(filter_noninstallable)

    if 'python_dependencies' in filter_names:
        filters.append(filter_python_dependencies)

    def filter_module(manifest):
        return all([check(manifest) for check in filters])

    return filter_module


def fast_search_manifests(path):
    filenames = ['__manifest__.py', '__openerp__.py']
    found_paths = []
    blacklist = ['setup', '.git']

    for manifest in filenames:
        manifest_path = path / manifest
        if manifest_path.exists():
            return [manifest_path]

    dirs_to_search = []
    if path.exists():
        for cpath in path.iterdir():
            if cpath.name in blacklist:
                continue
            if cpath.is_dir():
                dirs_to_search.append(cpath)
                continue
            # Never used because tested before looking up into dir
            # if cpath.name in filenames:
            #     found_paths.append(cpath)
            #     return
        else:
            for cpath in dirs_to_search:
                found_paths += fast_search_manifests(cpath)

    return found_paths


def find_modules(path, filters=None):
    modules = set()

    path = Path.cwd() / path

    manifest_globs = fast_search_manifests(path)

    check_module = get_filter(filters)

    for path in manifest_globs:
        manifest = Manifest.from_path(path)

        if not check_module(manifest):
            continue

        modules.add(manifest)

    return modules


def find_modules_paths(paths, filters=None, options=None):
    modules = set()

    if options and not options.exclude_odoo:
        odoo_path = base_addons_path()
        if odoo_path:
            paths.add(odoo_path)

    for path in paths:
        modules = modules.union(
            find_modules(Path(path), filters=filters)
        )

    return modules


def get_manifest(manifest, render_description=False):
    return Manifest.from_path(manifest, render_description)


def build_module_dependencies(
    modules,
    modules_lst=False,
    deps=None,
    quiet=True
):
    if deps is None:
        deps = {}

    to_process = modules_lst or []

    while len(to_process) > 0:
        cur_module = to_process.pop()
        if cur_module not in modules:
            continue

        dependencies = modules[cur_module].depends
        deps[cur_module] = set(dependencies)

        for dep in dependencies:
            if dep not in deps:
                to_process.append(dep)

            if not quiet and dep not in modules:
                print((
                    "Module {cur_module} depends on {dep} "
                    "which isn't in addons_path"
                ).format(cur_module=cur_module, dep=dep))

    return deps


def build_dependencies(
    modules,
    modules_lst,
    lookup_auto_install=True,
    deps=None,
    quiet=True
):
    deps = build_module_dependencies(
        modules,
        modules_lst,
        deps=deps,
        quiet=quiet
    )

    if not lookup_auto_install:
        return deps

    old_deps_length = 0

    while len(deps) != old_deps_length:
        old_deps_length = len(deps)

        to_install = []

        for name, module in modules.items():
            if not module.auto_install or name in deps:
                continue

            module_deps = module.depends
            for dep in module_deps:
                if dep not in deps:
                    break
            else:
                if len(module_deps) > 0:
                    to_install.append(name)

        deps = build_module_dependencies(
            modules,
            to_install,
            deps=deps,
            quiet=quiet
        )

    return deps


def base_addons_path():
    odoo_path = (
        module_path("odoo", raise_not_found=False) or
        module_path("openerp", raise_not_found=False)
    )
    return odoo_path


def find_addons_paths(paths, options=False):
    """
    Find addons in provided paths.

    Args:
      paths (List<Path>): A list of paths to search for addons.

    Returns:
      Set<Manifest>: A list of manifests
    """
    filters = set()
    filters.add('installable')

    if options and not options.exclude_odoo:
        odoo_path = base_addons_path()
        if odoo_path:
            paths.add(odoo_path)

    if options and options.include_odoo_entrypoints:
        entry_point = "odoo_addons_paths"
        for ep in pkg_resources.iter_entry_points(group=entry_point):
            functor = ep.load()
            new_paths = functor()
            paths |= new_paths

    modules = find_modules_paths(
        paths,
        filters=filters
    )

    found_paths = set()

    for manifest in modules:
        found_paths.add(manifest.path.parent)

    return found_paths



import giturlparse

from . import fields
from . import models


class ValidationError(Exception):
    pass


class Version(object):

    def __init__(self, since, until=None):
        self.since = since
        self.until = until


class KeyValue(models.Extendable, models.ParentedObject, models.BaseModel):
    def extend(self, other):
        options = self._data.copy()
        options.update(other._data if other and other._data else {})
        return KeyValue(options)

    def to_dict(self):
        return self._data.copy()


class RepoConfig(models.ParentedObject, models.Extendable, models.BaseModel):

    url = fields.Url(
        "Url of repository",
        availability=Version(since=1),
        default=None,
    )

    commit = fields.String(
        "Commit of the object",
        availability=Version(since=1),
        default=None,
    )

    branch = fields.String(
        "Branch of the object",
        availability=Version(since=1),
        default=None,
    )

    private_key = fields.String(
        default=None,
    )

    auth = fields.Boolean(
        default=None
    )

    @property
    def ref(self):
        service = self.get_parent(ServiceManifest)

        if (
            service and
            service.resolved.odoo and
            service.resolved.odoo.version
        ):
            default = service.resolved.odoo.version
        else:
            default = None

        return self.commit or self.branch or default

    @property
    def repo_path(self):
        url = giturlparse.parse(self.url.lower())

        if url.valid:
            path = "_".join([
                url.host.replace('.', '_'),
                url.owner.replace('/', '_'),
                url.repo.replace('/', '_')
            ])
        else:
            path = self.url

        return path

    def extend(self, other):
        if not other:
            return self

        return RepoConfig({
            "url": other.url or self.url,
            "commit": other.commit or self.commit,
            "branch": other.branch or self.branch,
            "private_key": other.private_key or self.private_key,
        })

    def to_dict(self):
        return {
            "url": self.url,
            "commit": self.commit,
            "branch": self.branch,
            "ref": self.ref,
            "private_key": self.private_key
        }


class OdooConfig(models.ParentedObject, models.Extendable, models.BaseModel):
    version = fields.String()

    repo = fields.Object(
        "Repository Configuration",
        object_class=RepoConfig,
        availability=Version(since=1)
    )

    options = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    def extend(self, other):
        if not other:
            return self

        return OdooConfig({
            "version": (
                other.version
                if other.version
                else self.version
            ),
            "repo": (
                other.repo.extend(self.repo)
                if other.repo
                else self.repo
            ),
            "options": (
                other.options.extend(self.options)
                if other.options
                else self.options
            )
        })

    def to_dict(self):
        return {
            "version": self.version,
            "repo": self.repo.to_dict() if self.repo else {},
            "options": self.options.to_dict() if self.options else {}
        }


class ManifestProxy(models.ParentedObject, models.BaseModel):
    reference = fields.String()

    @classmethod
    def parse(klass, data):
        return ManifestProxy({"reference": data})


class ServiceManifest(
    models.ParentedObject,
    models.Extendable,
    models.BaseModel
):
    """
    ServiceManifest
    ===============

    This object represent all properties that can be set
    on a manifest for odoo services.

    Properties:

    name (String): A string representing the name of the
        service configuration. For example, you may want to define
        a service staging and production with different settings.

    inherit (ServiceManifest): A reference to an other service manifest.
        It's possible to reference manifests by name. In that case, a service
        will be able to merge its own properties with a parent manifest
        configuration.

        For example, you may want to have a production server using the
        production branch of certain repositories. While a staging environment
        tracking the staging branch. To make things simpler, you could have all
        addons defined in the staging environment. But you could also define
        custom branches for production and keep the rest as the staging
        environment is using.

    odoo (OdooConfig): A field referencing an odoo configuration.

    addons (List<RepoConfig>): A list of repository configuration.

    labels (KeyValue): A key value store of labels.

    env (KeyValue): A key value store of environment variables.
    """
    name = fields.String(
        "name",
        availability=Version(since="1")
    )

    inherit = fields.ProxyObject(
        "Reference to parent manifest",
        getter="get_inherited_object",
        availability=Version(since="1")
    )

    odoo = fields.Object(
        "Odoo Configuration",
        object_class=OdooConfig,
        availability=Version(since="1")
    )

    addons = fields.Dict(
        "List of addons",
        object_class=RepoConfig,
        key="repo_path",
    )

    labels = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    env = fields.Object(
        "Odoo Custom Options",
        object_class=KeyValue,
        availability=Version(since=1)
    )

    def get_inherited_object(self, key):
        parent = self.get_parent(ServiceManifests)
        return parent.services[key]

    # TODO refactor to make auto extandable objects
    # Guess how they can be extended without specifying
    # properties.
    def resolved_prop(self, prop_name):
        if not self.inherit:
            return getattr(self, prop_name)

        other_val = self.inherit.resolved_prop(prop_name)
        self_val = getattr(self, prop_name)

        if not other_val:
            return self_val

        if not self_val:
            return other_val

        # TODO return Extendable Collections in field with list/dict
        if isinstance(other_val, models.Extendable):
            return other_val.extend(self_val)
        elif isinstance(other_val, dict):
            ret_val = other_val.copy()
            for key, value in self_val.items():
                if (
                    key in ret_val and
                    isinstance(ret_val[key], models.Extendable)
                ):
                    ret_val[key] = ret_val[key].extend(value)
                else:
                    ret_val[key] = value
            return ret_val
        elif isinstance(other_val, list):
            return other_val + ret_val
        else:
            print("whoa")

    @property
    def resolved(self):
        return self.extend(self.inherit)

    def extend(self, other):
        if not other:
            return self

        return ServiceManifest({
            "name": self.name,
            "odoo": self.resolved_prop('odoo'),
            "addons": self.resolved_prop('addons'),
            "labels": self.resolved_prop('labels'),
            "env": self.resolved_prop('env'),
        })

    # To dict should be guessable from properties defined on class
    def to_dict(self):
        inherit = (
            self._data['inherit']['ref']
            if 'inherit' in self._data
            else None
        )

        addons = self.addons or {}

        vals = {
            "name": self.name,
            "addons": {
                key: val.to_dict()
                for key, val in addons.items()
            }
        }

        if inherit:
            vals['inherit'] = inherit

        for attr in ['odoo', 'env', 'labels']:
            val = getattr(self, attr)
            if val:
                vals[attr] = val.to_dict()

        return vals


class ServiceManifests(models.BaseModel):
    services = fields.Dict(
        "List of services",
        object_class=ServiceManifest,
        key="name"
    )

    def to_dict(self):
        return {
            "services": {
                key: val.to_dict()
                for key, val in self.services.items()
            }
        }
